ScratchMedia subclass: #ImageMedia
	instanceVariableNames: 'form rotationCenter textBox jpegBytes compositeForm savedForm '
	classVariableNames: 'JPEGPlaceholderForm '
	poolDictionaries: ''
	category: 'Scratch-Objects'!
!ImageMedia commentStamp: '<historical>' prior: 0!
I am a Scratch image media object. I hold a still image that can be used as the costume for a Scratch object.

Instance variables:
	form			my form
	rotationCenter	my rotation center
	textBox			optional text layer
	jpegBytes		optional JPEG data for this image if it was compressed
	compositeForm	form that combines my base form with my text layer; same as form if no text

When there is no text layer and the image contains no transparent pixels, it can be compressed using JPEG. In that case, jpegBytes contains the compressed image and form is just a cache of decompressed image. If the image is not edited, the original jpegBytes are retained and saved with the project to avoid losing quality through repeated compress/decompress cycles.

When there is a text layer, compositeForm caches the base form with the text painted on top of it. This composite image is saved with the project to allow the Java player to use it. (Since the Java player does not have the same set of fonts as Scratch, it is not possible for it to recreate the compositeForm.)
!


!ImageMedia methodsFor: 'initialization' stamp: 'ee 8/11/2008 12:13'!
initialize

	super initialize.
	mediaName _ 'costume' localized, '1'.
	form _ (Form extent: 8@8 depth: 4) fillWhite.
	rotationCenter _ form extent // 2.	"relative to the top-left corner of the Form"
	textBox _ nil.
	jpegBytes _ nil.
	compositeForm _ nil.
! !


!ImageMedia methodsFor: 'accessing' stamp: 'jm 8/3/2008 23:37'!
compositeForm
	"Answer my composite form, which is my form with the text painted on top of it."

	compositeForm ifNotNil: [^ compositeForm].
	self hasTextLayer ifFalse: [^ compositeForm _ form].

	compositeForm _ Form extent: form extent depth: 32.
	form displayOn: compositeForm.
	textBox
		suppressAntiAliasing: true;
		drawTextOn: ((FormCanvas on: compositeForm) copyOffset: textBox topLeft).

	^ compositeForm
! !

!ImageMedia methodsFor: 'accessing' stamp: 'jm 5/3/2004 19:38'!
form

	^ form
! !

!ImageMedia methodsFor: 'accessing' stamp: 'CWB 7/14/2012 19:42'!
form: aForm

	form _ aForm ditherToBlackAndWhite.
	form couldBeColorReduced
		ifTrue: [form _ form colorReduced].  "reduce color depth to save space"

	rotationCenter _ form extent // 2.
	jpegBytes _ compositeForm _ nil.
! !

!ImageMedia methodsFor: 'accessing' stamp: 'jm 10/12/2007 13:57'!
hasTextLayer
	"Return true if I have a text layer."

	^ textBox notNil and: [textBox contents size > 0]
! !

!ImageMedia methodsFor: 'accessing' stamp: 'jm 6/3/2004 18:24'!
infoString
	"Answer a string for this media, typically something about its size."

	^ form width printString, 'x', form height printString
! !

!ImageMedia methodsFor: 'accessing' stamp: 'jm 6/30/2004 12:06'!
isImage
	"Answer true if I am an image media object."

	^ true
! !

!ImageMedia methodsFor: 'accessing' stamp: 'jm 10/12/2007 14:16'!
jpegBytes

	^ jpegBytes
! !

!ImageMedia methodsFor: 'accessing' stamp: 'jm 10/20/2007 20:35'!
mediaSizeInKilobytes
	"Returns the storage size of this media in kilobytes, rounded to the nearest kilobyte."

	| bytes |
	bytes _ 0.
	jpegBytes
		ifNil: [form hibernate. bytes _ form bits size]
		ifNotNil: [bytes _ jpegBytes size].
	self hasTextLayer ifTrue: [
		self compositeForm hibernate.
		bytes _ bytes + compositeForm bits size].

	bytes < 1024 ifTrue: [
		 ^ (bytes / 1024) roundTo: 0.01].

	^ (bytes + 512) // 1024
! !

!ImageMedia methodsFor: 'accessing' stamp: 'jm 5/12/2004 16:26'!
mediaType
	"Answer a string describing this type of media."

	^ 'image'
! !

!ImageMedia methodsFor: 'accessing' stamp: 'jm 5/3/2004 19:39'!
rotationCenter

	^ rotationCenter
! !

!ImageMedia methodsFor: 'accessing' stamp: 'jm 5/3/2004 19:39'!
rotationCenter: aPoint

	rotationCenter _ aPoint.
! !

!ImageMedia methodsFor: 'accessing' stamp: 'jm 10/18/2007 15:55'!
savedForm

	^ savedForm
! !

!ImageMedia methodsFor: 'accessing' stamp: 'jm 12/9/2005 15:41'!
textBox

	^ textBox
! !

!ImageMedia methodsFor: 'accessing' stamp: 'jm 12/9/2005 15:41'!
textBox: aMultilineStringMorphOrNil

	textBox _ aMultilineStringMorphOrNil.
! !

!ImageMedia methodsFor: 'accessing' stamp: 'jm 1/5/2006 20:21'!
thumbnailFormExtent: extent
	"Answer a Form of the given extent containing a thumbnail of me."

	| f compositeF scale shrunk |
	f _ (Form extent: extent depth: 16) fillColor: Color transparent.
	form ifNotNil: [
		compositeF _ self compositeForm.
		scale _ (f width / compositeF width) min: (f height / compositeF height).
		scale < 1
			ifTrue: [
				shrunk _ (compositeF asFormOfDepth: 16)
					magnify: compositeF boundingBox
					by: scale
					smoothing: 2]
			ifFalse: [
				shrunk _ compositeF].

		shrunk displayOn: f at: (f extent - shrunk extent) // 2 rule: Form paint].
	^ f
! !


!ImageMedia methodsFor: 'copying' stamp: 'jm 7/25/2006 10:24'!
copy
	"Return a copy of me with a copy of my textbox."

	^ self shallowCopy textBox: textBox copy
! !


!ImageMedia methodsFor: 'other' stamp: 'jm 5/19/2009 13:20'!
exportFilename: fileName for: stageOrSprite
	"Export my image to a file with the given name. Save sprite costumes as GIF's, dropping color resolution if necessary."

	| myForm f fName |
	myForm _ self compositeForm.

	"save as GIF if image has under 256 colors"
	((myForm depth <= 8) or: [myForm couldBeColorReduced]) ifTrue: [
		f _ myForm colorReduced8Bit.
		(fileName asLowercase endsWith: '.gif')
			ifTrue: [fName _ fileName]
			ifFalse: [fName _ fileName, '.gif'].
		GIFReadWriter putForm: f onFileNamed: fName.
		^ self].

	"otherwise, save as BMP to retain color resolution"
	(fileName asLowercase endsWith: '.bmp')
		ifTrue: [fName _ fileName]
		ifFalse: [fName _ fileName, '.bmp'].

	(stageOrSprite isKindOf: ScratchSpriteMorph)
		ifTrue: [
			f _ Form extent: myForm extent depth: 32.
			f fillColor: Color white.  "BMP cannot handle transparency; use white as BG color"
			myForm displayOn: f at: form offset negated rule: Form paint]
		ifFalse: [
			f _ myForm asFormOfDepth: 32].

	f writeBMPFileNamed: fileName.
! !

!ImageMedia methodsFor: 'other' stamp: 'jm 10/18/2007 16:44'!
jpegCompressIfPossibleQuality: quality
	"Attempt to compress my form using JPEG at the given quality level. If successful, save the compressed JPEG data in my jpegBytes instance variable. If JPEG compression does not yield a sufficient savings, then set jpegBytes to nil. If jpegBytes is non-nil when I am saved, then my form will not be saved and the form will be reconstructed from the JPEG data when this media is read back in. Answer true if JPEG compression was done."

	| oldSz tally |
	jpegBytes ifNotNil: [^ false].

	form unhibernate.
	(form isKindOf: ColorForm)
		ifTrue: [
			(form colorsUsed includes: Color transparent) ifTrue: [^ false]]
		ifFalse: [
			form depth > 1 ifTrue: [
				tally _ form tallyPixelValues.
				(tally at: 1) > 0 ifTrue: [^ false]]].  "contains transparent pixels"

	"no transparency; attempt jpeg compression"
	form hibernate.
	oldSz _ form bits size.
	jpegBytes _ FastJPEG compress: form quality: quality.
	jpegBytes size <= (oldSz // 3)
		ifFalse: [jpegBytes _ nil].  "not enough savings; don't use JPEG"

	^ jpegBytes notNil
! !

!ImageMedia methodsFor: 'other' stamp: 'jm 10/18/2007 15:47'!
jpegCompressIfPossibleQuality: quality saveOriginal: saveFlag
	"Attempt to compress my form using JPEG at the given quality level. if saveFlag is true and I was not previously compressed, save my original (pre-compression) image so that compression can be undone later (e.g. after uploading the project)."

	| wasCompressed |
	saveFlag & quality notNil
		ifTrue: [savedForm _ form]
		ifFalse: [savedForm _ nil].

	wasCompressed _ self jpegCompressIfPossibleQuality: quality.

	wasCompressed & saveFlag not ifTrue: [
		"show the result of doing the compression"
		form _ FastJPEG uncompress: jpegBytes.
		compositeForm _ nil].

	^ wasCompressed
! !

!ImageMedia methodsFor: 'other' stamp: 'jm 10/17/2007 13:06'!
revertToUncompressed
	"Revert to my original, uncompressed Form after compression with the 'saveOriginal' flag."

	savedForm ifNotNil: [
		form _ savedForm.
		compositeForm _ nil.
		jpegBytes _ nil.
		savedForm _ nil].
! !

!ImageMedia methodsFor: 'other' stamp: 'jm 10/18/2007 15:55'!
shareFormWith: anImageMedia
	"Share the same form object with the given image media object."

	form _ anImageMedia form.
	jpegBytes _ anImageMedia jpegBytes.
	savedForm _ anImageMedia savedForm.
! !


!ImageMedia methodsFor: 'object i/o' stamp: 'jm 7/25/2006 10:26'!
fieldsVersion

	^ 4
! !

!ImageMedia methodsFor: 'object i/o' stamp: 'jm 10/18/2007 15:45'!
initFieldsFrom: anObjStream version: classVersion

	| f |
	super initFieldsFrom: anObjStream version: classVersion.
	self initFieldsNamed: #(
		form
		rotationCenter 
	) from: anObjStream.

	(form isKindOf: ColorForm) ifTrue: [  "convert ColorForm to 32-bit color"
		f _ Form extent: form extent depth: 32.
		form displayOn: f.
		form _ f].

	compositeForm _ jpegBytes _ nil.
	classVersion = 1 ifTrue: [^ self].

	"fields added in version 2"
	self initFieldsNamed: #(
		textBox
	) from: anObjStream.

	classVersion = 2 ifTrue: [^ self].

	"fields added in version 3"
	self initFieldsNamed: #(
		jpegBytes
	) from: anObjStream.
	jpegBytes ifNotNil: [form _ FastJPEG uncompress: jpegBytes].

	classVersion = 3 ifTrue: [^ self].

	"fields added in version 4"
	self initFieldsNamed: #(
		compositeForm
	) from: anObjStream.
! !

!ImageMedia methodsFor: 'object i/o' stamp: 'jm 8/4/2008 07:29'!
storeFieldsOn: anObjStream

	| oldForm f |
	super storeFieldsOn: anObjStream.

	oldForm _ form.

	compositeForm _ nil.
	self hasTextLayer ifTrue: [
		compositeForm _ self compositeForm hibernate.
		compositeForm couldBeColorReduced ifTrue: [
			f _ compositeForm colorReduced hibernate.
			(f bits size < compositeForm bits size) ifTrue: [compositeForm _ f]]].

	jpegBytes
		ifNil: [form _ form hibernate]
		ifNotNil: [form _ JPEGPlaceholderForm].

	self storeFieldsNamed: #(
		form
		rotationCenter
		textBox
		jpegBytes
		compositeForm
	) on: anObjStream.

	form _ oldForm.
	compositeForm _ nil.
! !
